package moe.gensoukyo.haloshield.client.model

import net.minecraft.client.model.ModelBase
import net.minecraft.client.model.ModelRenderer
import net.minecraftforge.fml.relauncher.Side
import net.minecraftforge.fml.relauncher.SideOnly
import java.lang.reflect.Field
import java.util.*
import kotlin.collections.ArrayList

/**
 * 从一个模型类中提取所有 [ModelRenderer] 信息
 * 并存储
 */
@SideOnly(Side.CLIENT)
object ModelPartFinder {
    /**
     * 存放一个模型类所有可能的部件字段
     * 包括普通的 ModelRenderer 字段和 ModelRenderer[] 数组字段
     */
    private val models = object: ClassValue<ModelFieldInfo>() {
        val BOUND = ModelBase::class.java
        override fun computeValue(type: Class<*>): ModelFieldInfo {
            val fields = type.declaredFields

            val plains = ArrayList<Field>(fields.size)
            val arrays = ArrayList<Field>(fields.size)

            fields.filter {
                it.type() != TYPE_INVALID
            }.onEach {
                if (it.isAccessible) {
                    return@onEach
                }
                try {
                    it.isAccessible = true
                } catch (ignored: Exception) {
                    // 无访问权限的field会被删除掉
                }
            }.forEach {
                when (it.type()) {
                    TYPE_MODEL -> plains.add(it)
                    TYPE_MODEL_ARRAY -> arrays.add(it)
                    else -> throw InternalError("Unknown Model Field!")
                }
            }

            val parentClass = type.superclass
            val superman: ModelFieldInfo =
                if (!BOUND.isAssignableFrom(parentClass)) ModelFieldInfo.EMPTY
                else get(parentClass)

            return ModelFieldInfo(
                plains + superman.plainFields,
                arrays + superman.arrayFields
            )
        }
    }

    class ModelFieldInfo(
        val plainFields: List<Field>,
        val arrayFields: List<Field>
    ) {
        companion object {
            val EMPTY = ModelFieldInfo(emptyList(), emptyList())
        }
    }

    fun <M: ModelBase> getFields(clazz: Class<M>) = models.get(clazz)!!

    @Suppress("UNCHECKED_CAST")
    fun findParts(model: ModelBase): List<ModelRenderer> {
        val info = models.get(model::class.java)
        val result = ArrayList<ModelRenderer>(
            info.plainFields.size + info.arrayFields.size * 6)
        info.plainFields.forEach {
            result.add(it.get(model) as ModelRenderer)
        }
        info.arrayFields.forEach {
            val arr = it.get(model) as Array<ModelRenderer>
            result.addAll(arr)
        }
        return result
    }

    private const val TYPE_INVALID = 0
    private const val TYPE_MODEL = 1
    private const val TYPE_MODEL_ARRAY = 2

    private val MODEL_TOP_TYPE = ModelRenderer::class.java
    private fun Field.type(): Int {
        return when {
            // ModelRenderer 引用
            MODEL_TOP_TYPE.isAssignableFrom(this.type) -> TYPE_MODEL
            // ModelRenderer 数组
            this.type.isArray && MODEL_TOP_TYPE.isAssignableFrom(this.type.componentType) -> TYPE_MODEL_ARRAY
            else -> TYPE_INVALID
        }
    }
}