package moe.gensoukyo.haloshield.client.renderer

import moe.gensoukyo.haloshield.SpartanShieldsMod
import moe.gensoukyo.haloshield.api.client.IShieldRenderControl
import moe.gensoukyo.haloshield.client.caps.IllegalModelException
import moe.gensoukyo.haloshield.client.compat.cnpc.AbstractNpcEnergyShield
import moe.gensoukyo.haloshield.client.model.EnergyShieldModel
import moe.gensoukyo.haloshield.common.util.EventBusHalter
import moe.gensoukyo.haloshield.core.client.api.IRenderPreparationAccessor
import net.minecraft.client.Minecraft
import net.minecraft.client.model.ModelBase
import net.minecraft.client.renderer.GlStateManager
import net.minecraft.client.renderer.entity.RenderLivingBase
import net.minecraft.client.renderer.entity.RenderManager
import net.minecraft.entity.EntityLivingBase
import net.minecraft.util.ResourceLocation
import net.minecraftforge.fml.relauncher.Side
import net.minecraftforge.fml.relauncher.SideOnly
import org.lwjgl.opengl.GL11

/**
 * @throws IllegalModelException 有的时候模型类型转换失败，此时抛出异常
 */
@SideOnly(Side.CLIENT)
open class RenderEnergyShield<E: EntityLivingBase> private constructor(
    renderManagerIn: RenderManager,
    private val entityType: Class<E>,
    shieldModel: ModelBase
)  : RenderLivingBase<E>(renderManagerIn, shieldModel, 0.0f) {
    internal constructor(entityType: Class<E>, shieldModel: ModelBase)
            : this(DEFAULT_RENDER_MANAGER, entityType, shieldModel)

    companion object {
        val EMPTY_RENDERER = object : RenderEnergyShield<EntityLivingBase>
            (EntityLivingBase::class.java, EnergyShieldModel.EMPTY_MODEL) {
            override fun doRender(
                entity: EntityLivingBase,
                x: Double,
                y: Double,
                z: Double,
                entityYaw: Float,
                partialTicks: Float
            ) {}

            override fun preRenderCallback(
                entity: EntityLivingBase,
                partialTick: Float
            ) {}

            override fun applyRotations(
                entity: EntityLivingBase,
                ageInTicks: Float,
                rotationYaw: Float,
                partialTicks: Float
            ) {}

            override fun toString(): String {
                return "Empty ${RenderEnergyShield::class.qualifiedName}"
            }
        }

        private val SHIELD_TEXTURE = ResourceLocation(
            SpartanShieldsMod.MODID,
            "textures/entity/shield.png"
        )
        private val DEFAULT_RENDER_MANAGER: RenderManager = Minecraft.getMinecraft().renderManager

    }

    // 渲染状态传入

    private var tickExisted: Int = 0
    internal open fun tellMeTickExisted(tickIn: Int): RenderEnergyShield<E> {
        tickExisted = tickIn
        return this
    }

    /**
     * NPC下
     * 第一次渲染时会先给 [AbstractNpcEnergyShield] 绑定控制器
     * 然后再生成 NPC 引用的实体的渲染器
     * 导致 NPC 引用实体的渲染器的 [controller] 控制器为null
     * 所以这个值为可空类型
     */
    private var controller: IShieldRenderControl? = null
    internal open fun bindController(c: IShieldRenderControl): RenderEnergyShield<E> {
        controller = c
        return this
    }

    // 渲染代码

    /**
     * 防止护盾显示名字
     */
    override fun canRenderName(entity: E) = false

    private var origRendererMixin: IRenderPreparationAccessor<E>? = null

    /**
     * 这个方法可以让渲染器使用其他实体的旋转设置
     *
     * 此数值不为空时
     * [applyRotations] 将调用这个函数以取代原渲染器的 [applyRotations] 方法
     * 该函数在被调用一次后会变被清空为 null
     */
    var rotOverride: ((E, Float, Float, Float) -> Unit)? = null
    var preparationOverride: ((E, Float) -> Unit)? = null

    @Suppress("UNCHECKED_CAST")
    override fun preRenderCallback(entity: E, partialTick: Float) {
        if (preparationOverride != null) {
            preparationOverride!!(entity, partialTick)
            preparationOverride = null
        } else {
            origRendererMixin?.callPreRenderCallback(entity, partialTick)
        }
        partialRecord = partialTick

        setupOGL(partialTick)
    }

    private var partialRecord = 0f

    @Suppress("UNCHECKED_CAST")
    override fun applyRotations(entity: E, ageInTicks: Float, rotationYaw: Float, partialTicks: Float) {
        if (rotOverride != null) {
            rotOverride!!(entity, ageInTicks, rotationYaw, partialTicks)
            rotOverride = null
            return
        }
        if (origRendererMixin == null) {
            origRendererMixin = Minecraft.getMinecraft().renderManager
                .getEntityClassRenderObject<E>(entityType) as? IRenderPreparationAccessor<E>
        }
        origRendererMixin?.callApplyRotations(entity, ageInTicks, rotationYaw, partialTicks)
        //super.applyRotations(entity, ageInTicks, rotationYaw, partialTicks)
    }

    override fun doRender(entity: E, x: Double, y: Double, z: Double, entityYaw: Float, partialTicks: Float) {
        if (controller == null || controller!!.visible.not()) {
            return
        }
        EventBusHalter.pause()
        super.doRender(entity, x, y, z, entityYaw, partialTicks)
        EventBusHalter.resume()
        unsetOGL()
    }

    override fun renderModel(
        entitylivingbaseIn: E,
        limbSwing: Float,
        limbSwingAmount: Float,
        ageInTicks: Float,
        netHeadYaw: Float,
        headPitch: Float,
        scaleFactor: Float
    ) {
        super.renderModel(
            entitylivingbaseIn,
            limbSwing,
            limbSwingAmount,
            ageInTicks,
            netHeadYaw,
            headPitch,
            scaleFactor
        )
    }

    private fun setupOGL(partialTick: Float) {
        // Moving Texture
        GlStateManager.matrixMode(GL11.GL_TEXTURE)
        GlStateManager.loadIdentity()

        val moveSpeed = 0.04f
        val texScale = 2f
        val offset = tickExisted + partialTick
        GlStateManager.translate(moveSpeed * offset, moveSpeed * offset, 0f)
        GlStateManager.scale(texScale, texScale, 0f)
        GlStateManager.matrixMode(GL11.GL_MODELVIEW)

        // Blend Mode: Add
        GlStateManager.enableBlend()
        GlStateManager.blendFunc(
            GlStateManager.SourceFactor.SRC_ALPHA,
            GlStateManager.DestFactor.ONE
        )
        setupDynamicOglData(partialTick)
    }

    private fun setupDynamicOglData(partialTick: Float) {
        val c = controller?.getColor(partialTick) ?: return
        GlStateManager.color(c.x, c.y, c.z, c.w)
    }


    private fun unsetOGL() {
        GlStateManager.matrixMode(GL11.GL_TEXTURE)
        GlStateManager.loadIdentity()
        GlStateManager.matrixMode(GL11.GL_MODELVIEW)
        GlStateManager.disableBlend()
    }
    override fun getEntityTexture(entity: E) = SHIELD_TEXTURE
}